/*
 * Copyright (C) 2011 Google Inc.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

package org.ros.rosjava_tutorial_pubsub;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;

import java.util.TimerTask;

import javax.imageio.ImageIO;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JSeparator;
import javax.swing.Timer;

import org.apache.commons.logging.Log;
import org.ros.concurrent.CancellableLoop;
import org.ros.message.MessageListener;
import org.ros.namespace.GraphName;
import org.ros.node.AbstractNodeMain;
import org.ros.node.ConnectedNode;
import org.ros.node.NodeMain;
import org.ros.node.topic.Publisher;
import org.ros.node.topic.Subscriber;

/**
 * A simple {@link Subscriber} {@link NodeMain}.
 * 
 * @author damonkohler@google.com (Damon Kohler)
 */
public class Listener extends AbstractNodeMain implements KeyListener, ItemListener, ActionListener {

	// various
	public static final int LEFT = 0;
	public static final int RIGHT = 1;
	public int calibrationCounter=0;
	
	// motion
	private int lSpeed=0, rSpeed=0;
	public static final int MAX_SPEED = 350; // mm/s
	public static final int SPEED_STEP = 26;
	public static final int ROTATION_STEP = 5;
	public int currentSpeed=0;
	public int currentSpeedIndex=0;
	public int currentRotation=0;
	public int currentRotationIndex=0;
	public static final int NO_ROTATION = 0;
	public static final int ROTATE_LEFT = 1;
	public static final int ROTATE_RIGHT = 2;
	public int pivotType = NO_ROTATION;
	public int pivotCounter = 0;	
	Timer timer;
	
	// UI
	private byte flagStatus = 0;
	private byte batteryValue = 0;
	private byte proxValues [] = new byte[4];
	private byte groundValues [] = new byte[4];
	private short desiredVel [] = new short[2];
	JLabel proxLabels [] = {new JLabel("prox0: "), new JLabel("prox1: "), new JLabel("prox2 "), new JLabel("prox3: ")};
	JLabel proxLabelValues [] = {new JLabel("0"), new JLabel("0"), new JLabel("0"), new JLabel("0")};
	JProgressBar proxBars [] = {new JProgressBar(JProgressBar.VERTICAL, 0, 255), new JProgressBar(JProgressBar.VERTICAL, 0, 255), new JProgressBar(JProgressBar.VERTICAL, 0, 255), new JProgressBar(JProgressBar.VERTICAL, 0, 255)};
	JLabel groundLabels [] = {new JLabel("ground0: "), new JLabel("ground1: "), new JLabel("ground2 "), new JLabel("ground3: ")};
	JLabel groundLabelValues [] = {new JLabel("0"), new JLabel("0"), new JLabel("0"), new JLabel("0")};
	JProgressBar groundBars [] = {new JProgressBar(JProgressBar.VERTICAL, 0, 255), new JProgressBar(JProgressBar.VERTICAL, 0, 255), new JProgressBar(JProgressBar.VERTICAL, 0, 255), new JProgressBar(JProgressBar.VERTICAL, 0, 255)};
	JLabel flagsLabelValues [] = {new JLabel("0"), new JLabel("0")};
	JLabel motorsLabelValues [] = {new JLabel("0"), new JLabel("0")};
	JLabel currSpeedLabel = new JLabel(String.valueOf(currentSpeed));
	JCheckBox onOffController;
	JCheckBox onOffMicroAcc;
	JCheckBox onOffObstacleAvoidance;
	JCheckBox onOffCliffAvoidance;
	JCheckBox onOffAutonomous;
	JLabel imageLabel = new JLabel("");
	JPanel panel2;
	JButton moveFw, moveBw, moveRight, moveLeft, moveStop;
	JProgressBar batteryBar = new JProgressBar(0, 100);
	JButton calibrateSensor;
	
	// behavior
	private boolean isAutonomous = false;
	Timer autonomousTimer;
	private int OBSTACLE_THR = 20;
	private int OBSTACLE_SPEED = 20;
	private int SAME_DIR_COUNT = 20;	// based on 20 ms task base time
	private int OBSTACLE_TOO_NEAR_THR = 125;
	private int maintainDirCounter = 0;
	private int sumSensorsX=0, sumSensorsY=0; 
	private int sumLeft=0, sumRight=0;
	private int sumLeftPrev=0, sumRightPrev=0;
	private int baseSpeed=OBSTACLE_SPEED;
	private int OA_ROT_RIGHT = 0;
	private int OA_ROT_LEFT = 1;
	private int OA_FORWARD = 2;
	private int rotationDir = OA_ROT_RIGHT;
	
	
	// ROS
	public Subscriber<std_msgs.UInt8MultiArray> subscribProximity;
	public Subscriber<std_msgs.UInt8MultiArray> subscribGround;
	public Subscriber<std_msgs.UInt8> subscribFlags;
	public Subscriber<sensor_msgs.CompressedImage> subscribImage;
	public Publisher<std_msgs.Int16MultiArray> motorsPub;
	public Publisher<std_msgs.Int8> flagsPub;
	
	
	ActionListener pivotingTask = new ActionListener() {
		public void actionPerformed(ActionEvent evt) {
			if(pivotType != NO_ROTATION) {
				
				if(pivotCounter == 0) {
								
					lSpeed = 0;
					rSpeed = 0;
					pivotType = NO_ROTATION;
					timer.stop();
								
				} else {
							
					if(pivotCounter > 0) {
						pivotCounter--;
						lSpeed = 0;
						rSpeed = 20;
					} else {
						pivotCounter++;
						lSpeed = 20;
						rSpeed = 0;
					}
				
				}
							
			}				
		}
	};	

	ActionListener autonomousTask = new ActionListener() {
		public void actionPerformed(ActionEvent evt) {

//			sumSensorsX = -(((proxValues[0]&0xFF)*8)/10) - ((proxValues[1]&0xFF)>>1) - ((proxValues[2]&0xFF)>>1) - (((proxValues[3]&0xFF)*8)/10); // + ((rand()%60)-30);
//			sumSensorsY = ((proxValues[0]&0xFF)>>1) - (((proxValues[1]&0xFF)*8)/10) + (((proxValues[2]&0xFF)*8)/10) - ((proxValues[3]&0xFF)>>1); // + ((rand()%60)-30);
//			
//			rSpeed = OBSTACLE_SPEED + ((OBSTACLE_SPEED * (sumSensorsX + sumSensorsY))>>3);
//			lSpeed = OBSTACLE_SPEED + ((OBSTACLE_SPEED * (sumSensorsX - sumSensorsY))>>3);
//
//			motorsLabelValues[0].setText(String.valueOf(lSpeed));
//			motorsLabelValues[1].setText(String.valueOf(rSpeed)); 
//			
////			if(rSpeed > MAX_SPEED) {
////				rSpeed = MAX_SPEED;
////			}
////			if(rSpeed < 0) {
////				rSpeed = 0;
////			}
////			if(lSpeed > MAX_SPEED) {
////				lSpeed = MAX_SPEED;
////			}
////			if(lSpeed < 0) {
////				lSpeed = 0;
////			}
//			
//			if(rSpeed > MAX_SPEED) {
//				rSpeed = MAX_SPEED;
//			}
//			if(rSpeed < -MAX_SPEED) {
//				rSpeed = -MAX_SPEED;
//			}
//			if(lSpeed > MAX_SPEED) {
//				lSpeed = MAX_SPEED;
//			}
//			if(lSpeed < -MAX_SPEED) {
//				lSpeed = -MAX_SPEED;
//			}			
			
			if(maintainDirCounter>0) {
				maintainDirCounter--;
				if(rotationDir == OA_ROT_RIGHT) {
					if((proxValues[0]&0xFF)>OBSTACLE_TOO_NEAR_THR || (proxValues[1]&0xFF)>OBSTACLE_TOO_NEAR_THR) {
						lSpeed = baseSpeed;
						rSpeed = -baseSpeed;							
					}
				}
				if(rotationDir == OA_ROT_LEFT) {
					if((proxValues[2]&0xFF)>OBSTACLE_TOO_NEAR_THR || (proxValues[3]&0xFF)>OBSTACLE_TOO_NEAR_THR) {
						lSpeed = -baseSpeed;
						rSpeed = baseSpeed;
					}
				}
			} else {
				lSpeed = baseSpeed;
				rSpeed = baseSpeed;
				rotationDir = OA_FORWARD;
				sumLeftPrev = sumLeft;
				sumRightPrev = sumRight;
				sumLeft = (proxValues[0]&0xFF)+(proxValues[1]&0xFF);
				sumRight = (proxValues[2]&0xFF)+(proxValues[3]&0xFF);
				if(sumLeft > sumRight) {
					//if(sumLeft > (sumRight-sumLeft)) { //OBSTACLE_THR) {	
					if(sumLeft > OBSTACLE_THR-(sumLeft-sumLeftPrev)) {
						if((proxValues[0]&0xFF)>OBSTACLE_TOO_NEAR_THR || (proxValues[1]&0xFF)>OBSTACLE_TOO_NEAR_THR) {
							lSpeed = baseSpeed; //((proxValues[0]&0xFF)+(proxValues[1]&0xFF)); //OBSTACLE_SPEED;
							rSpeed = -baseSpeed;							
						} else {
							lSpeed = baseSpeed; //((proxValues[0]&0xFF)+(proxValues[1]&0xFF)); //OBSTACLE_SPEED;
							rSpeed = 0;
						}
						maintainDirCounter=SAME_DIR_COUNT;
						rotationDir = OA_ROT_RIGHT;
					}
				} else {
					//if(sumRight > (sumLeft-sumRight)) { //OBSTACLE_THR) {
					if(sumRight > OBSTACLE_THR-(sumRight-sumRightPrev)) {
						if((proxValues[2]&0xFF)>OBSTACLE_TOO_NEAR_THR || (proxValues[3]&0xFF)>OBSTACLE_TOO_NEAR_THR) {
							lSpeed = -baseSpeed;
							rSpeed = baseSpeed;
						} else {
							lSpeed = 0;
							rSpeed = baseSpeed; //((proxValues[2]&0xFF)+(proxValues[3]&0xFF)); //OBSTACLE_SPEED;
						}
						maintainDirCounter=SAME_DIR_COUNT;
						rotationDir = OA_ROT_LEFT;
					}
				}
				motorsLabelValues[0].setText(String.valueOf(lSpeed));
				motorsLabelValues[1].setText(String.valueOf(rSpeed)); 
			}

		}
	};	
	
	
	@Override
	public GraphName getDefaultNodeName() {
		return new GraphName("rosjava_tutorial_pubsub/listener");
	}

  @Override
  public void onStart(ConnectedNode connectedNode) {
    final Log log = connectedNode.getLog();
    
    subscribProximity = connectedNode.newSubscriber("proximity", std_msgs.UInt8MultiArray._TYPE);
    subscribProximity.addMessageListener(new MessageListener<std_msgs.UInt8MultiArray>() {
      @Override
      public void onNewMessage(std_msgs.UInt8MultiArray message) {
    	  proxValues = message.getData();
    	  //System.out.println(proxValues[0] + ", " + proxValues[1] + ", " + proxValues[2] + ", " + proxValues[3]);    	  
    	  proxBars[0].setValue(proxValues[0]&0xFF);
    	  proxBars[1].setValue(proxValues[1]&0xFF);
    	  proxBars[2].setValue(proxValues[2]&0xFF);
    	  proxBars[3].setValue(proxValues[3]&0xFF);
    	  
    	  if(calibrationCounter<3) {
    		  calibrationCounter++;
    	  } else {
    		  flagStatus &= ~(1 << 4);	// do calibration only once
    		  flagsLabelValues[1].setText(String.valueOf(flagStatus));  
    	  }
      }
    });
    
    subscribGround = connectedNode.newSubscriber("ground", std_msgs.UInt8MultiArray._TYPE);
    subscribGround.addMessageListener(new MessageListener<std_msgs.UInt8MultiArray>() {
      @Override
      public void onNewMessage(std_msgs.UInt8MultiArray message) {
    	  groundValues = message.getData();
    	  //System.out.println(groundValues[0] + ", " + groundValues[1] + ", " + groundValues[2] + ", " + groundValues[3]);
    	  groundBars[0].setValue(groundValues[3]&0xFF);
    	  groundBars[1].setValue(groundValues[2]&0xFF);
    	  groundBars[2].setValue(groundValues[1]&0xFF);
    	  groundBars[3].setValue(groundValues[0]&0xFF);      
      }
    });    
    
    subscribFlags = connectedNode.newSubscriber("battery", std_msgs.UInt8._TYPE);
    subscribFlags.addMessageListener(new MessageListener<std_msgs.UInt8>() {
      @Override
      public void onNewMessage(std_msgs.UInt8 message) {
    	  batteryValue = message.getData();
    	  batteryBar.setValue(batteryValue);
      }
    }); 
    
    subscribImage = connectedNode.newSubscriber("camera/image/compressed", sensor_msgs.CompressedImage._TYPE);
    subscribImage.addMessageListener(new MessageListener<sensor_msgs.CompressedImage>() {
      @Override
      public void onNewMessage(sensor_msgs.CompressedImage message) {
    	  InputStream in = new ByteArrayInputStream(message.getData());
    	  try {
    		  BufferedImage bImage = ImageIO.read(in);
    		  //System.out.println("height = " + bImage.getHeight());
    		  //System.out.println("width = " + bImage.getWidth());
    		  //imageLabel.setSize(bImage.getWidth(), bImage.getHeight());
    		  
    		  //imageLabel.setIcon(new ImageIcon(bImage));
    		  
    		  ImageIcon ii = new ImageIcon(bImage);
    		  imageLabel.setIcon(new RotatedIcon(ii, RotatedIcon.Rotate.DOWN));

    		    //panel2.getGraphics().drawImage(bImage, 0, 0, null);
    		    
  		  } catch (IOException e) {
  			  System.out.println(e.getMessage());
		  }
    	  
      }
    }); 
    //subscribImage.setQueueLimit(100);
    
    final class WheelPhoneTop extends JPanel {
    	public void paint(Graphics g) {
    		Graphics2D g2 = (Graphics2D) g;
    		g2.setPaint(Color.lightGray);
    		g2.fillRect(5, 70, 140, 140);
    		g2.fillOval(5, 0, 140, 140);
    		g2.setPaint(Color.BLACK);
    		g2.fillRect(35, 125, 80, 30);
    		g2.fillOval(0, 90, 10, 100);
    		g2.fillOval(140, 90, 10, 100);
    	}
    }
    
    final class WheelPhoneFront extends JPanel {
    	public void paint(Graphics g) {
    		Graphics2D g2 = (Graphics2D) g;
    		g2.setPaint(Color.lightGray);
    		g2.fillRect(5, 50, 140, 70);
    		g2.fillOval(5, 85, 140, 70);
    		g2.setPaint(Color.BLACK);
    		g2.fillRect(35, 10, 80, 90);
    		g2.fillOval(0, 35, 10, 100);
    		g2.fillOval(140, 35, 10, 100);
    	}
    }
    
    WheelPhoneTop wpt = new WheelPhoneTop();
   
    WheelPhoneFront wpf = new WheelPhoneFront();
    
    //Create and set up the window.
    JFrame frame = new JFrame("Sensors");
    frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
    frame.setSize(380, 830);
    frame.addWindowListener(new WindowAdapter() {
    	public void windowClosing(WindowEvent evt) {
    		onExit();
    	}
    });
    
    //GridLayout gridLayout = new GridLayout(0,2);
    //GridLayout moveLayout = new GridLayout(3,3);
    
    final JPanel movePanel = new JPanel();
    //movePanel.setLayout(moveLayout);
    
    final JPanel panel = new JPanel();
    //panel.setLayout(gridLayout);
    panel.setLayout(null);

    JLabel logoTitle = new JLabel();
    logoTitle.setIcon(new ImageIcon("wheelphone-font.jpg"));
    logoTitle.setBounds(40, 10, 300, 80);
    panel.add(logoTitle);
    
    JLabel lblProximity = new JLabel("Proximity sensors");
    lblProximity.setBounds(30, 100, 150, 20);
    panel.add(lblProximity);
    
    JLabel lblGrounds = new JLabel("Ground sensors");
    lblGrounds.setBounds(220, 100, 150, 20);
    panel.add(lblGrounds);
    
    int k=0;
    for(k=0; k<4; k++) {
        proxBars[k].setStringPainted(true);
        proxBars[k].setValue(0);
        groundBars[k].setStringPainted(true);
        groundBars[k].setValue(0);
    }
    
	
	proxBars[0].setBounds(33, 140, 14, 50);
	proxBars[1].setBounds(63, 120, 14, 50);
	proxBars[2].setBounds(93, 120, 14, 50);
	proxBars[3].setBounds(123, 140, 14, 50);
	
	groundBars[0].setBounds(198, 270, 14, 50);
	groundBars[1].setBounds(248, 290, 14, 50);
	groundBars[2].setBounds(288, 290, 14, 50);
	groundBars[3].setBounds(338, 270, 14, 50);
	
    for(k=0; k<4; k++) {
        panel.add(proxBars[k]);
        panel.add(groundBars[k]);
    }
    
    wpt.setBounds(10, 180, 160, 220);
    panel.add(wpt);
    
    wpf.setBounds(200, 130, 160, 160);
    panel.add(wpf);   
    
    JButton b = new JButton();


    
    calibrateSensor = new JButton();
    calibrateSensor.setBounds(185, 360, 190, 40);
    calibrateSensor.addKeyListener(this);
    calibrateSensor.addActionListener(this);     
    calibrateSensor.setLayout(new BorderLayout());
    JLabel label1 = new JLabel("Calibrate sensors");
    JLabel label2 = new JLabel("and reset odometry");
    calibrateSensor.add(BorderLayout.NORTH,label1);
    calibrateSensor.add(BorderLayout.SOUTH,label2);
    panel.add(calibrateSensor);
    
    JLabel lblBattery = new JLabel("Battery: ");
    lblBattery.setBounds(10, 430, 70, 20);
    panel.add(lblBattery);
    batteryBar.setBounds(90, 430, 200, 20);
    batteryBar.setStringPainted(true);
    panel.add(batteryBar);
    
    JLabel lblFlags = new JLabel("Flags: ");
    lblFlags.setBounds(10, 460, 70, 20);
    panel.add(lblFlags);
    flagsLabelValues[1].setBounds(90, 460, 100, 20);
    panel.add(flagsLabelValues[1]);
    
    onOffController = new JCheckBox("Speed controller");
    onOffController.addItemListener(this);
    onOffController.setBounds(10, 490, 170, 20);
    panel.add(onOffController);
    onOffController.addKeyListener(this);

    onOffMicroAcc = new JCheckBox("Soft acceleration");
    onOffMicroAcc.addItemListener(this);
    onOffMicroAcc.setBounds(190, 490, 170, 20);
    panel.add(onOffMicroAcc);
    onOffMicroAcc.addKeyListener(this);
    
    onOffObstacleAvoidance = new JCheckBox("Obstacle avoidance");
    onOffObstacleAvoidance.addItemListener(this);
    onOffObstacleAvoidance.setBounds(10, 520, 170, 20);
    panel.add(onOffObstacleAvoidance);
    onOffObstacleAvoidance.addKeyListener(this);
    
    onOffCliffAvoidance = new JCheckBox("Cliff avoidance");
    onOffCliffAvoidance.addItemListener(this);
    onOffCliffAvoidance.setBounds(190, 520, 170, 20);
    panel.add(onOffCliffAvoidance);
    onOffCliffAvoidance.addKeyListener(this);
    
    JLabel lblLeftSpeed = new JLabel("Left: ");
    lblLeftSpeed.setBounds(130, 550, 50, 20);
    panel.add(lblLeftSpeed);
    motorsLabelValues[0].setBounds(180, 550, 100, 20);
    panel.add(motorsLabelValues[0]);
    
    JLabel lblRightSpeed = new JLabel("Right: ");
    lblRightSpeed.setBounds(250, 550, 50, 20);
    panel.add(lblRightSpeed);
    motorsLabelValues[1].setBounds(300, 550, 50, 20);
    panel.add(motorsLabelValues[1]);
    
    moveFw = new JButton();
    moveFw.setIcon(new ImageIcon(("up-arrow.png")));
    moveFw.setBounds(70, 580, 50, 50);
    moveFw.addKeyListener(this);
    moveFw.addActionListener(this); 
    moveBw = new JButton();
    moveBw.setIcon(new ImageIcon("down-arrow.png"));
    moveBw.setBounds(70, 700, 50, 50);
    moveBw.addKeyListener(this);
    moveBw.addActionListener(this); 
    moveRight = new JButton();
    moveRight.setIcon(new ImageIcon("right-arrow.png"));
    moveRight.setBounds(130, 640, 50, 50);
    moveRight.addKeyListener(this);
    moveRight.addActionListener(this); 
    moveLeft = new JButton();
    moveLeft.setIcon(new ImageIcon("left-arrow.png"));
    moveLeft.setBounds(10, 640, 50, 50);
    moveLeft.addKeyListener(this);
    moveLeft.addActionListener(this); 
    moveStop = new JButton();
    moveStop.setIcon(new ImageIcon("stop.png"));
    moveStop.setBounds(70, 640, 50, 50);
    moveStop.addKeyListener(this);
    moveStop.addActionListener(this);

    panel.add(moveFw);
    panel.add(moveLeft);
    panel.add(moveStop);
    panel.add(moveRight);
    panel.add(moveBw);
    
    JLabel logo = new JLabel();
    logo.setIcon(new ImageIcon("WheelphoneLogo.png"));
    logo.setBounds(200, 580, 150, 150);
    panel.add(logo);
    
    onOffAutonomous = new JCheckBox("Autonomous");
    onOffAutonomous.addItemListener(this);
    onOffAutonomous.setBounds(10, 770, 170, 20);
    panel.add(onOffAutonomous);
    onOffAutonomous.addKeyListener(this);
    
    frame.getContentPane().add(panel);
    
    //Display the window.
    //frame.pack();
    frame.setVisible(true);
    
    frame.addKeyListener(this);  
          
    JFrame imageFrame = new JFrame("Image");
    imageFrame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
    imageFrame.setSize(480, 800);
    
    panel2 = new JPanel();
    //panel2.setSize(480, 320);

    GridLayout gridLayout = new GridLayout(1,1);
    panel2.setLayout(gridLayout);
    panel2.add(imageLabel);
    
    imageFrame.getContentPane().add(panel2);
    //imageFrame.pack();
    imageFrame.setVisible(true);
    imageFrame.setLocation(frame.getX()+frame.getWidth(), frame.getY());

    
    motorsPub = connectedNode.newPublisher("motorSpeed", std_msgs.Int16MultiArray._TYPE);
        // This CancellableLoop will be canceled automatically when the node shuts
        // down.
        
        connectedNode.executeCancellableLoop(new CancellableLoop() {

          @Override
          protected void setup() {
        	  desiredVel[LEFT] = (short) lSpeed;
        	  desiredVel[RIGHT] = (short) rSpeed;
          }

          @Override
          protected void loop() throws InterruptedException {
        	  	desiredVel[LEFT] = (short) lSpeed;
        	  	desiredVel[RIGHT] = (short) rSpeed;
  	        	std_msgs.Int16MultiArray value = motorsPub.newMessage();
  	        	value.setData(desiredVel);
  	        	motorsPub.publish(value);
  	        	Thread.sleep(200);
          }
        });    
        
        flagsPub = connectedNode.newPublisher("flagStatus", std_msgs.Int8._TYPE);
        // This CancellableLoop will be canceled automatically when the node shuts
        // down.
        
        connectedNode.executeCancellableLoop(new CancellableLoop() {

          @Override
          protected void setup() {
          }

          @Override
          protected void loop() throws InterruptedException {
              	std_msgs.Int8 value = flagsPub.newMessage();
              	value.setData(flagStatus);
              	flagsPub.publish(value);
              	Thread.sleep(200);
          }
        });                 
        
        timer = new Timer(50, pivotingTask);	// 50 ms
        autonomousTimer = new Timer(20, autonomousTask);	// 20 ms
  }

  
	@Override
	public void keyPressed(KeyEvent e) {
		// TODO Auto-generated method stub
		//System.out.println("pressed");
		if(e.getKeyCode() == KeyEvent.VK_LEFT) {			
			if(currentSpeedIndex == 0) {    				
    			pivotType = ROTATE_LEFT;
    			pivotCounter += 40;
    			if(!timer.isRunning()) {
    				timer.start();
    			}      			
			} else {    			
    			if(Math.abs(currentSpeedIndex) < 3) {
    				if(currentRotationIndex < 13) {
    					currentRotationIndex++;
    				} else {
    					currentRotationIndex = 13;
    				}
    			} else {
    				if(currentSpeedIndex >= 0) {    				
			    		if(currentRotationIndex < (13-(currentSpeedIndex/3))) {
			    			currentRotationIndex += currentSpeedIndex/3;
			    		} else {
			    			currentRotationIndex = 13;
			    		}		
    				} else {
			    		if(currentRotationIndex < (13+(currentSpeedIndex/3))) {
			    			currentRotationIndex -= currentSpeedIndex/3;
			    		} else {
			    			currentRotationIndex = 13;
			    		}    					
    				}
    			}
			}				
		} else if(e.getKeyCode() == KeyEvent.VK_RIGHT) {
			if(currentSpeedIndex == 0) {    				
    			pivotType = ROTATE_RIGHT;
    			pivotCounter -= 40;	// about 45 degrees CW   	
    			if(!timer.isRunning()) {
    				timer.start();
    			}      			
			} else {    			
    			if(Math.abs(currentSpeedIndex) < 3) {
		    		if(currentRotationIndex > -13) {
		    			currentRotationIndex--;
		    		} else {
		    			currentRotationIndex = -13;
		    		}    				
    			} else {  	
    				if(currentSpeedIndex >= 0) {   
			    		if(currentRotationIndex > (-13+(currentSpeedIndex/3))) {
			    			currentRotationIndex -= currentSpeedIndex/3;
			    		} else {
			    			currentRotationIndex = -13;
			    		}  
    				} else {
			    		if(currentRotationIndex > (-13-(currentSpeedIndex/3))) {
			    			currentRotationIndex += currentSpeedIndex/3;
			    		} else {
			    			currentRotationIndex = -13;
			    		}      					
    				}
    			}
			}				
		} else if(e.getKeyCode() == KeyEvent.VK_UP) {
			if(currentRotationIndex == 0) {
				if(currentSpeedIndex < 13) {	//13*SPEED_STEP=130 (max is 127)
					currentSpeedIndex++;
				} else {
					currentSpeedIndex = 13;
				}    				
			} else {
				currentRotationIndex = 0;
			}
			pivotCounter = 0;
			pivotType = NO_ROTATION;     
		} else if(e.getKeyCode() == KeyEvent.VK_DOWN) {
			if(currentRotationIndex == 0) {
				if(currentSpeedIndex > -13) {
					currentSpeedIndex--;
				} else {
					currentSpeedIndex = -13;
				}    				
			} else {
				currentRotationIndex = 0;
			}
			pivotCounter = 0;
			pivotType = NO_ROTATION; 			
		} else if(e.getKeyCode() == KeyEvent.VK_ENTER) {
			currentSpeedIndex = 0;
			currentRotationIndex = 0;	
			pivotCounter = 0;
			pivotType = NO_ROTATION;  			
		}
		
		if(currentSpeedIndex >= 0) {	// forward direction
    		rSpeed = currentSpeedIndex + currentRotationIndex;
    		lSpeed = currentSpeedIndex - currentRotationIndex;
		} else {	// backward direction
    		rSpeed = currentSpeedIndex - currentRotationIndex;	
    		lSpeed = currentSpeedIndex + currentRotationIndex;
		}		
		
		rSpeed *= SPEED_STEP;
		lSpeed *= SPEED_STEP;
		
		if(rSpeed > MAX_SPEED) {
			rSpeed = MAX_SPEED;
		}
		if(rSpeed < -MAX_SPEED) {
			rSpeed = -MAX_SPEED;
		}
		if(lSpeed > MAX_SPEED) {
			lSpeed = MAX_SPEED;
		}
		if(lSpeed < -MAX_SPEED) {
			lSpeed = -MAX_SPEED;
		}			
		
  	  	motorsLabelValues[0].setText(String.valueOf(lSpeed));
  	  	motorsLabelValues[1].setText(String.valueOf(rSpeed)); 
  	  	
  	  	baseSpeed = lSpeed;
		
	}
	
	@Override
	public void keyReleased(KeyEvent arg0) {
		// TODO Auto-generated method stub		
	}
	
	@Override
	public void keyTyped(KeyEvent e) {
		// TODO Auto-generated method stub
	}
	
    /** Listens to the check boxes. */
    public void itemStateChanged(ItemEvent e) {
   	
        Object source = e.getItemSelectable();
 
        if (source == onOffController) {
            if (e.getStateChange() == ItemEvent.DESELECTED) {
            	flagStatus &= ~(1 << 0);
            	//System.out.println("controller deselected");
            } else {
            	flagStatus |= (1 << 0);
            	//System.out.println("controller selected");
            }
        } else if (source == onOffMicroAcc) {
            if (e.getStateChange() == ItemEvent.DESELECTED) {
            	flagStatus &= ~(1 << 1);
            	//System.out.println("acc deselected");
            } else {
            	flagStatus |= (1 << 1);
            	//System.out.println("acc selected");
            }        	
        } else if (source == onOffObstacleAvoidance) {
            if (e.getStateChange() == ItemEvent.DESELECTED) {
            	flagStatus &= ~(1 << 2);
            	//System.out.println("acc deselected");
            } else {
            	flagStatus |= (1 << 2);
            	//System.out.println("acc selected");
            }        	
        } else if (source == onOffCliffAvoidance) {
            if (e.getStateChange() == ItemEvent.DESELECTED) {
            	flagStatus &= ~(1 << 3);
            	//System.out.println("acc deselected");
            } else {
            	flagStatus |= (1 << 3);
            	//System.out.println("acc selected");
            }        	
        } else if (source == onOffAutonomous) {
            if (e.getStateChange() == ItemEvent.DESELECTED) {
            	isAutonomous = false;
            	if(autonomousTimer.isRunning()) {
            		autonomousTimer.stop();
            	}
            	lSpeed = 0;
            	rSpeed = 0;
            	motorsLabelValues[0].setText(String.valueOf(lSpeed));
          	  	motorsLabelValues[1].setText(String.valueOf(rSpeed)); 
            } else {
            	isAutonomous = true;
    			if(!autonomousTimer.isRunning()) {
    				autonomousTimer.start();
    			} 
            }        	
        } 
        
  	  	flagsLabelValues[1].setText(String.valueOf(flagStatus));      
 
    }

	@Override
	public void actionPerformed(ActionEvent ae) {
		// TODO Auto-generated method stub
		if(ae.getSource() == moveFw) {
			if(currentRotationIndex == 0) {
				if(currentSpeedIndex < 13) {	//13*SPEED_STEP=130 (max is 127)
					currentSpeedIndex++;
				} else {
					currentSpeedIndex = 13;
				}    				
			} else {
				currentRotationIndex = 0;
			}
			pivotCounter = 0;
			pivotType = NO_ROTATION;    			
		} else if(ae.getSource() == moveBw) {
			if(currentRotationIndex == 0) {
				if(currentSpeedIndex > -13) {
					currentSpeedIndex--;
				} else {
					currentSpeedIndex = -13;
				}    				
			} else {
				currentRotationIndex = 0;
			}
			pivotCounter = 0;
			pivotType = NO_ROTATION; 			

		} else if(ae.getSource() == moveRight) {
			if(currentSpeedIndex == 0) {    				
    			pivotType = ROTATE_RIGHT;
    			pivotCounter -= 40;	// 45 degrees CW 
    			if(!timer.isRunning()) {
    				timer.start();
    			}   			
			} else {    			
    			if(Math.abs(currentSpeedIndex) < 3) {
		    		if(currentRotationIndex > -13) {
		    			currentRotationIndex--;
		    		} else {
		    			currentRotationIndex = -13;
		    		}    				
    			} else {  	
    				if(currentSpeedIndex >= 0) {   
			    		if(currentRotationIndex > (-13+(currentSpeedIndex/3))) {
			    			currentRotationIndex -= currentSpeedIndex/3;
			    		} else {
			    			currentRotationIndex = -13;
			    		}  
    				} else {
			    		if(currentRotationIndex > (-13-(currentSpeedIndex/3))) {
			    			currentRotationIndex += currentSpeedIndex/3;
			    		} else {
			    			currentRotationIndex = -13;
			    		}      					
    				}
    			}
			}			
		} else if(ae.getSource() == moveLeft) {
			if(currentSpeedIndex == 0) {    				
    			pivotType = ROTATE_LEFT;
    			pivotCounter += 40;
    			if(!timer.isRunning()) {    				
    				timer.start();
    			} 
			} else {    			
    			if(Math.abs(currentSpeedIndex) < 3) {
    				if(currentRotationIndex < 13) {
    					currentRotationIndex++;
    				} else {
    					currentRotationIndex = 13;
    				}
    			} else {
    				if(currentSpeedIndex >= 0) {    				
			    		if(currentRotationIndex < (13-(currentSpeedIndex/3))) {
			    			currentRotationIndex += currentSpeedIndex/3;
			    		} else {
			    			currentRotationIndex = 13;
			    		}		
    				} else {
			    		if(currentRotationIndex < (13+(currentSpeedIndex/3))) {
			    			currentRotationIndex -= currentSpeedIndex/3;
			    		} else {
			    			currentRotationIndex = 13;
			    		}    					
    				}
    			}
			}			
		} else if(ae.getSource() == moveStop) {		
			currentSpeedIndex = 0;
			currentRotationIndex = 0;	
			pivotCounter = 0;
			pivotType = NO_ROTATION;   			
			
		} else if(ae.getSource() == calibrateSensor) {
			calibrationCounter = 0;
			flagStatus |= (1 << 4);
			flagsLabelValues[1].setText(String.valueOf(flagStatus));  
		}
		
		if(currentSpeedIndex >= 0) {	// forward direction
    		rSpeed = currentSpeedIndex + currentRotationIndex;
    		lSpeed = currentSpeedIndex - currentRotationIndex;
		} else {	// backward direction
    		rSpeed = currentSpeedIndex - currentRotationIndex;	
    		lSpeed = currentSpeedIndex + currentRotationIndex;
		}		

			
		rSpeed *= SPEED_STEP;			
		lSpeed *= SPEED_STEP;
		
		if(rSpeed > MAX_SPEED) {
			rSpeed = MAX_SPEED;
		}
		if(rSpeed < -MAX_SPEED) {
			rSpeed = -MAX_SPEED;
		}
		if(lSpeed > MAX_SPEED) {
			lSpeed = MAX_SPEED;
		}
		if(lSpeed < -MAX_SPEED) {
			lSpeed = -MAX_SPEED;
		}		
		
  	  	motorsLabelValues[0].setText(String.valueOf(lSpeed));
  	  	motorsLabelValues[1].setText(String.valueOf(rSpeed));  
  	  	
  	  	baseSpeed = lSpeed;
		
	}	
	
	public void onExit() {
		subscribProximity.shutdown();
		subscribGround.shutdown();
		subscribFlags.shutdown();
		subscribImage.shutdown();
		motorsPub.shutdown();
		flagsPub.shutdown();
		//System.err.println("Exit");
		System.exit(0);
	}
	
}
